# coding: utf-8

import json
import uuid

from stix2.canonicalization.Canonicalize import canonicalize


class Vulnerability:
    """Main Vulnerability class for OpenCTI

    Manages vulnerability information including CVE data in the OpenCTI platform.

    :param opencti: instance of :py:class:`~pycti.api.opencti_api_client.OpenCTIApiClient`
    :type opencti: OpenCTIApiClient
    """

    def __init__(self, opencti):
        """Initialize the Vulnerability instance.

        :param opencti: OpenCTI API client instance
        :type opencti: OpenCTIApiClient
        """
        self.opencti = opencti
        self.properties = """
            id
            standard_id
            entity_type
            parent_types
            spec_version
            created_at
            updated_at
            status {
                id
                template {
                  id
                  name
                  color
                }
            }
            createdBy {
                ... on Identity {
                    id
                    standard_id
                    entity_type
                    parent_types
                    spec_version
                    identity_class
                    name
                    description
                    roles
                    contact_information
                    x_opencti_aliases
                    created
                    modified
                    objectLabel {
                        id
                        value
                        color
                    }
                }
                ... on Organization {
                    x_opencti_organization_type
                    x_opencti_reliability
                }
                ... on Individual {
                    x_opencti_firstname
                    x_opencti_lastname
                }
            }
            objectOrganization {
                id
                standard_id
                name
            }
            objectMarking {
                id
                standard_id
                entity_type
                definition_type
                definition
                created
                modified
                x_opencti_order
                x_opencti_color
            }
            objectLabel {
                id
                value
                color
            }
            externalReferences {
                edges {
                    node {
                        id
                        standard_id
                        entity_type
                        source_name
                        description
                        url
                        hash
                        external_id
                        created
                        modified
                        importFiles {
                            edges {
                                node {
                                    id
                                    name
                                    size
                                    metaData {
                                        mimetype
                                        version
                                    }
                                }
                            }
                        }
                    }
                }
            }
            revoked
            confidence
            created
            modified
            name
            description
            x_opencti_aliases
            x_opencti_cvss_vector_string
            x_opencti_cvss_base_score
            x_opencti_cvss_base_severity
            x_opencti_cvss_attack_vector
            x_opencti_cvss_attack_complexity
            x_opencti_cvss_privileges_required
            x_opencti_cvss_user_interaction
            x_opencti_cvss_scope
            x_opencti_cvss_confidentiality_impact
            x_opencti_cvss_integrity_impact
            x_opencti_cvss_availability_impact
            x_opencti_cvss_exploit_code_maturity
            x_opencti_cvss_remediation_level
            x_opencti_cvss_report_confidence
            x_opencti_cvss_temporal_score
            x_opencti_cvss_v2_vector_string
            x_opencti_cvss_v2_base_score
            x_opencti_cvss_v2_access_vector
            x_opencti_cvss_v2_access_complexity
            x_opencti_cvss_v2_authentication
            x_opencti_cvss_v2_confidentiality_impact
            x_opencti_cvss_v2_integrity_impact
            x_opencti_cvss_v2_availability_impact
            x_opencti_cvss_v2_exploitability
            x_opencti_cvss_v2_remediation_level
            x_opencti_cvss_v2_report_confidence
            x_opencti_cvss_v2_temporal_score
            x_opencti_cvss_v4_vector_string
            x_opencti_cvss_v4_base_score
            x_opencti_cvss_v4_base_severity
            x_opencti_cvss_v4_attack_vector
            x_opencti_cvss_v4_attack_complexity
            x_opencti_cvss_v4_attack_requirements
            x_opencti_cvss_v4_privileges_required
            x_opencti_cvss_v4_user_interaction
            x_opencti_cvss_v4_confidentiality_impact_v
            x_opencti_cvss_v4_confidentiality_impact_s
            x_opencti_cvss_v4_integrity_impact_v
            x_opencti_cvss_v4_integrity_impact_s
            x_opencti_cvss_v4_availability_impact_v
            x_opencti_cvss_v4_availability_impact_s
            x_opencti_cvss_v4_exploit_maturity
            x_opencti_cwe
            x_opencti_first_seen_active
            x_opencti_cisa_kev
            x_opencti_epss_score
            x_opencti_epss_percentile
            x_opencti_score
            importFiles {
                edges {
                    node {
                        id
                        name
                        size
                        metaData {
                            mimetype
                            version
                        }
                        objectMarking {
                            id
                            standard_id
                            entity_type
                            definition_type
                            definition
                            created
                            modified
                            x_opencti_order
                            x_opencti_color
                        }
                    }
                }
            }
        """

    @staticmethod
    def generate_id(name):
        """Generate a STIX ID for a Vulnerability.

        :param name: The name of the vulnerability (e.g., CVE ID)
        :type name: str
        :return: STIX ID for the vulnerability
        :rtype: str
        """
        name = name.lower().strip()
        data = {"name": name}
        data = canonicalize(data, utf8=False)
        id = str(uuid.uuid5(uuid.UUID("00abedb4-aa42-466c-9c01-fed23315a9b7"), data))
        return "vulnerability--" + id

    @staticmethod
    def generate_id_from_data(data):
        """Generate a STIX ID from vulnerability data.

        :param data: Dictionary containing 'name' key
        :type data: dict
        :return: STIX ID for the vulnerability
        :rtype: str
        """
        return Vulnerability.generate_id(data["name"])

    def list(self, **kwargs):
        """List Vulnerability objects.

        :param filters: the filters to apply
        :type filters: dict
        :param search: the search keyword
        :type search: str
        :param first: return the first n rows from the after ID (or the beginning if not set)
        :type first: int
        :param after: ID of the first row for pagination
        :type after: str
        :param orderBy: field to order results by
        :type orderBy: str
        :param orderMode: ordering mode (asc/desc)
        :type orderMode: str
        :param customAttributes: custom attributes to return
        :type customAttributes: str
        :param getAll: whether to retrieve all results
        :type getAll: bool
        :param withPagination: whether to include pagination info
        :type withPagination: bool
        :return: List of Vulnerability objects
        :rtype: list
        """
        filters = kwargs.get("filters", None)
        search = kwargs.get("search", None)
        first = kwargs.get("first", 100)
        after = kwargs.get("after", None)
        order_by = kwargs.get("orderBy", None)
        order_mode = kwargs.get("orderMode", None)
        custom_attributes = kwargs.get("customAttributes", None)
        get_all = kwargs.get("getAll", False)
        with_pagination = kwargs.get("withPagination", False)

        self.opencti.app_logger.info(
            "Listing Vulnerabilities with filters", {"filters": json.dumps(filters)}
        )
        query = (
            """
            query Vulnerabilities($filters: FilterGroup, $search: String, $first: Int, $after: ID, $orderBy: VulnerabilitiesOrdering, $orderMode: OrderingMode) {
                vulnerabilities(filters: $filters, search: $search, first: $first, after: $after, orderBy: $orderBy, orderMode: $orderMode) {
                    edges {
                        node {
                            """
            + (custom_attributes if custom_attributes is not None else self.properties)
            + """
                        }
                    }
                    pageInfo {
                        startCursor
                        endCursor
                        hasNextPage
                        hasPreviousPage
                        globalCount
                    }
                }
            }
        """
        )
        result = self.opencti.query(
            query,
            {
                "filters": filters,
                "search": search,
                "first": first,
                "after": after,
                "orderBy": order_by,
                "orderMode": order_mode,
            },
        )

        if get_all:
            final_data = []
            data = self.opencti.process_multiple(result["data"]["vulnerabilities"])
            final_data = final_data + data
            while result["data"]["vulnerabilities"]["pageInfo"]["hasNextPage"]:
                after = result["data"]["vulnerabilities"]["pageInfo"]["endCursor"]
                self.opencti.app_logger.debug(
                    "Listing Vulnerabilities", {"after": after}
                )
                result = self.opencti.query(
                    query,
                    {
                        "filters": filters,
                        "search": search,
                        "first": first,
                        "after": after,
                        "orderBy": order_by,
                        "orderMode": order_mode,
                    },
                )
                data = self.opencti.process_multiple(result["data"]["vulnerabilities"])
                final_data = final_data + data
            return final_data
        else:
            return self.opencti.process_multiple(
                result["data"]["vulnerabilities"], with_pagination
            )

    def read(self, **kwargs):
        """Read a Vulnerability object.

        :param id: the id of the Vulnerability
        :type id: str
        :param filters: the filters to apply if no id provided
        :type filters: dict
        :param customAttributes: custom attributes to return
        :type customAttributes: str
        :return: Vulnerability object
        :rtype: dict or None
        """
        id = kwargs.get("id", None)
        filters = kwargs.get("filters", None)
        custom_attributes = kwargs.get("customAttributes", None)
        if id is not None:
            self.opencti.app_logger.info("Reading Vulnerability", {"id": id})
            query = (
                """
                query Vulnerability($id: String!) {
                    vulnerability(id: $id) {
                        """
                + (
                    custom_attributes
                    if custom_attributes is not None
                    else self.properties
                )
                + """
                    }
                }
             """
            )
            result = self.opencti.query(query, {"id": id})
            return self.opencti.process_multiple_fields(result["data"]["vulnerability"])
        elif filters is not None:
            result = self.list(filters=filters)
            if len(result) > 0:
                return result[0]
            else:
                return None
        else:
            self.opencti.app_logger.error(
                "[opencti_vulnerability] Missing parameters: id or filters"
            )
            return None

    def create(self, **kwargs):
        """Create a Vulnerability object.

        :param name: the name of the Vulnerability (required)
        :type name: str
        :param stix_id: (optional) the STIX ID
        :type stix_id: str
        :param createdBy: (optional) the author ID
        :type createdBy: str
        :param objectMarking: (optional) list of marking definition IDs
        :type objectMarking: list
        :param objectLabel: (optional) list of label IDs
        :type objectLabel: list
        :param externalReferences: (optional) list of external reference IDs
        :type externalReferences: list
        :param revoked: (optional) whether the vulnerability is revoked
        :type revoked: bool
        :param confidence: (optional) confidence level (0-100)
        :type confidence: int
        :param lang: (optional) language
        :type lang: str
        :param created: (optional) creation date
        :type created: str
        :param modified: (optional) modification date
        :type modified: str
        :param description: (optional) description
        :type description: str
        :param x_opencti_aliases: (optional) list of aliases
        :type x_opencti_aliases: list
        :param x_opencti_cvss_vector_string: (optional) CVSS v3 vector string
        :type x_opencti_cvss_vector_string: str
        :param x_opencti_cvss_base_score: (optional) CVSS v3 base score
        :type x_opencti_cvss_base_score: float
        :param x_opencti_cvss_base_severity: (optional) CVSS v3 base severity
        :type x_opencti_cvss_base_severity: str
        :param x_opencti_cwe: (optional) CWE ID
        :type x_opencti_cwe: str
        :param x_opencti_cisa_kev: (optional) CISA KEV flag
        :type x_opencti_cisa_kev: bool
        :param x_opencti_epss_score: (optional) EPSS score
        :type x_opencti_epss_score: float
        :param x_opencti_epss_percentile: (optional) EPSS percentile
        :type x_opencti_epss_percentile: float
        :param x_opencti_score: (optional) OpenCTI score
        :type x_opencti_score: int
        :param x_opencti_first_seen_active: (optional) first seen active date
        :type x_opencti_first_seen_active: str
        :param x_opencti_stix_ids: (optional) list of additional STIX IDs
        :type x_opencti_stix_ids: list
        :param objectOrganization: (optional) list of organization IDs
        :type objectOrganization: list
        :param x_opencti_workflow_id: (optional) workflow ID
        :type x_opencti_workflow_id: str
        :param x_opencti_modified_at: (optional) custom modification date
        :type x_opencti_modified_at: str
        :param update: (optional) whether to update if exists (default: False)
        :type update: bool
        :param files: (optional) list of File objects to attach
        :type files: list
        :param filesMarkings: (optional) list of lists of marking definition IDs for each file
        :type filesMarkings: list
        :return: Vulnerability object
        :rtype: dict or None
        """
        stix_id = kwargs.get("stix_id", None)
        created_by = kwargs.get("createdBy", None)
        object_marking = kwargs.get("objectMarking", None)
        object_label = kwargs.get("objectLabel", None)
        external_references = kwargs.get("externalReferences", None)
        revoked = kwargs.get("revoked", None)
        confidence = kwargs.get("confidence", None)
        lang = kwargs.get("lang", None)
        created = kwargs.get("created", None)
        modified = kwargs.get("modified", None)
        name = kwargs.get("name", None)
        description = kwargs.get("description", None)
        x_opencti_aliases = kwargs.get("x_opencti_aliases", None)
        # CVSS3
        x_opencti_cvss_vector_string = kwargs.get("x_opencti_cvss_vector_string", None)
        x_opencti_cvss_base_score = kwargs.get("x_opencti_cvss_base_score", None)
        x_opencti_cvss_base_severity = kwargs.get("x_opencti_cvss_base_severity", None)
        x_opencti_cvss_attack_vector = kwargs.get("x_opencti_cvss_attack_vector", None)
        x_opencti_cvss_attack_complexity = kwargs.get(
            "x_opencti_cvss_attack_complexity", None
        )
        x_opencti_cvss_privileges_required = kwargs.get(
            "x_opencti_cvss_privileges_required", None
        )
        x_opencti_cvss_user_interaction = kwargs.get(
            "x_opencti_cvss_user_interaction", None
        )
        x_opencti_cvss_scope = kwargs.get("x_opencti_cvss_scope", None)
        x_opencti_cvss_confidentiality_impact = kwargs.get(
            "x_opencti_cvss_confidentiality_impact", None
        )
        x_opencti_cvss_integrity_impact = kwargs.get(
            "x_opencti_cvss_integrity_impact", None
        )
        x_opencti_cvss_availability_impact = kwargs.get(
            "x_opencti_cvss_availability_impact", None
        )
        x_opencti_cvss_exploit_code_maturity = kwargs.get(
            "x_opencti_cvss_exploit_code_maturity", None
        )
        x_opencti_cvss_remediation_level = kwargs.get(
            "x_opencti_cvss_remediation_level", None
        )
        x_opencti_cvss_report_confidence = kwargs.get(
            "x_opencti_cvss_report_confidence", None
        )
        x_opencti_cvss_temporal_score = kwargs.get(
            "x_opencti_cvss_temporal_score", None
        )
        # CVSS2
        x_opencti_cvss_v2_vector_string = kwargs.get(
            "x_opencti_cvss_v2_vector_string", None
        )
        x_opencti_cvss_v2_base_score = kwargs.get("x_opencti_cvss_v2_base_score", None)
        x_opencti_cvss_v2_access_vector = kwargs.get(
            "x_opencti_cvss_v2_access_vector", None
        )
        x_opencti_cvss_v2_access_complexity = kwargs.get(
            "x_opencti_cvss_v2_access_complexity", None
        )
        x_opencti_cvss_v2_authentication = kwargs.get(
            "x_opencti_cvss_v2_authentication", None
        )
        x_opencti_cvss_v2_confidentiality_impact = kwargs.get(
            "x_opencti_cvss_v2_confidentiality_impact", None
        )
        x_opencti_cvss_v2_integrity_impact = kwargs.get(
            "x_opencti_cvss_v2_integrity_impact", None
        )
        x_opencti_cvss_v2_availability_impact = kwargs.get(
            "x_opencti_cvss_v2_availability_impact", None
        )
        x_opencti_cvss_v2_exploitability = kwargs.get(
            "x_opencti_cvss_v2_exploitability", None
        )
        x_opencti_cvss_v2_remediation_level = kwargs.get(
            "x_opencti_cvss_v2_remediation_level", None
        )
        x_opencti_cvss_v2_report_confidence = kwargs.get(
            "x_opencti_cvss_v2_report_confidence", None
        )
        x_opencti_cvss_v2_temporal_score = kwargs.get(
            "x_opencti_cvss_v2_temporal_score", None
        )
        # CVSS4
        x_opencti_cvss_v4_vector_string = kwargs.get(
            "x_opencti_cvss_v4_vector_string", None
        )
        x_opencti_cvss_v4_base_score = kwargs.get("x_opencti_cvss_v4_base_score", None)
        x_opencti_cvss_v4_base_severity = kwargs.get(
            "x_opencti_cvss_v4_base_severity", None
        )
        x_opencti_cvss_v4_attack_vector = kwargs.get(
            "x_opencti_cvss_v4_attack_vector", None
        )
        x_opencti_cvss_v4_attack_complexity = kwargs.get(
            "x_opencti_cvss_v4_attack_complexity", None
        )
        x_opencti_cvss_v4_attack_requirements = kwargs.get(
            "x_opencti_cvss_v4_attack_requirements", None
        )
        x_opencti_cvss_v4_privileges_required = kwargs.get(
            "x_opencti_cvss_v4_privileges_required", None
        )
        x_opencti_cvss_v4_user_interaction = kwargs.get(
            "x_opencti_cvss_v4_user_interaction", None
        )
        x_opencti_cvss_v4_confidentiality_impact_v = kwargs.get(
            "x_opencti_cvss_v4_confidentiality_impact_v", None
        )
        x_opencti_cvss_v4_confidentiality_impact_s = kwargs.get(
            "x_opencti_cvss_v4_confidentiality_impact_s", None
        )
        x_opencti_cvss_v4_integrity_impact_v = kwargs.get(
            "x_opencti_cvss_v4_integrity_impact_v", None
        )
        x_opencti_cvss_v4_integrity_impact_s = kwargs.get(
            "x_opencti_cvss_v4_integrity_impact_s", None
        )
        x_opencti_cvss_v4_availability_impact_v = kwargs.get(
            "x_opencti_cvss_v4_availability_impact_v", None
        )
        x_opencti_cvss_v4_availability_impact_s = kwargs.get(
            "x_opencti_cvss_v4_availability_impact_s", None
        )
        x_opencti_cvss_v4_exploit_maturity = kwargs.get(
            "x_opencti_cvss_v4_exploit_maturity", None
        )
        # Others
        x_opencti_cwe = kwargs.get("x_opencti_cwe", None)
        x_opencti_cisa_kev = kwargs.get("x_opencti_cisa_kev", None)
        x_opencti_epss_score = kwargs.get("x_opencti_epss_score", None)
        x_opencti_epss_percentile = kwargs.get("x_opencti_epss_percentile", None)
        x_opencti_score = kwargs.get("x_opencti_score", None)
        x_opencti_first_seen_active = kwargs.get("x_opencti_first_seen_active", None)
        x_opencti_stix_ids = kwargs.get("x_opencti_stix_ids", None)
        granted_refs = kwargs.get("objectOrganization", None)
        x_opencti_workflow_id = kwargs.get("x_opencti_workflow_id", None)
        x_opencti_modified_at = kwargs.get("x_opencti_modified_at", None)
        update = kwargs.get("update", False)
        files = kwargs.get("files", None)
        files_markings = kwargs.get("filesMarkings", None)
        no_trigger_import = kwargs.get("noTriggerImport", None)
        embedded = kwargs.get("embedded", None)
        upsert_operations = kwargs.get("upsert_operations", None)

        if name is not None:
            self.opencti.app_logger.info("Creating Vulnerability", {"name": name})
            query = """
                mutation VulnerabilityAdd($input: VulnerabilityAddInput!) {
                    vulnerabilityAdd(input: $input) {
                        id
                        standard_id
                        entity_type
                        parent_types
                    }
                }
            """
            result = self.opencti.query(
                query,
                {
                    "input": {
                        "stix_id": stix_id,
                        "createdBy": created_by,
                        "objectMarking": object_marking,
                        "objectLabel": object_label,
                        "objectOrganization": granted_refs,
                        "externalReferences": external_references,
                        "revoked": revoked,
                        "confidence": confidence,
                        "lang": lang,
                        "created": created,
                        "modified": modified,
                        "name": name,
                        "description": description,
                        "x_opencti_aliases": x_opencti_aliases,
                        # CVSS3
                        "x_opencti_cvss_vector_string": x_opencti_cvss_vector_string,
                        "x_opencti_cvss_base_score": x_opencti_cvss_base_score,
                        "x_opencti_cvss_base_severity": x_opencti_cvss_base_severity,
                        "x_opencti_cvss_attack_vector": x_opencti_cvss_attack_vector,
                        "x_opencti_cvss_attack_complexity": x_opencti_cvss_attack_complexity,
                        "x_opencti_cvss_privileges_required": x_opencti_cvss_privileges_required,
                        "x_opencti_cvss_user_interaction": x_opencti_cvss_user_interaction,
                        "x_opencti_cvss_scope": x_opencti_cvss_scope,
                        "x_opencti_cvss_confidentiality_impact": x_opencti_cvss_confidentiality_impact,
                        "x_opencti_cvss_integrity_impact": x_opencti_cvss_integrity_impact,
                        "x_opencti_cvss_availability_impact": x_opencti_cvss_availability_impact,
                        "x_opencti_cvss_exploit_code_maturity": x_opencti_cvss_exploit_code_maturity,
                        "x_opencti_cvss_remediation_level": x_opencti_cvss_remediation_level,
                        "x_opencti_cvss_report_confidence": x_opencti_cvss_report_confidence,
                        "x_opencti_cvss_temporal_score": x_opencti_cvss_temporal_score,
                        # CVSS2
                        "x_opencti_cvss_v2_vector_string": x_opencti_cvss_v2_vector_string,
                        "x_opencti_cvss_v2_base_score": x_opencti_cvss_v2_base_score,
                        "x_opencti_cvss_v2_access_vector": x_opencti_cvss_v2_access_vector,
                        "x_opencti_cvss_v2_access_complexity": x_opencti_cvss_v2_access_complexity,
                        "x_opencti_cvss_v2_authentication": x_opencti_cvss_v2_authentication,
                        "x_opencti_cvss_v2_confidentiality_impact": x_opencti_cvss_v2_confidentiality_impact,
                        "x_opencti_cvss_v2_integrity_impact": x_opencti_cvss_v2_integrity_impact,
                        "x_opencti_cvss_v2_availability_impact": x_opencti_cvss_v2_availability_impact,
                        "x_opencti_cvss_v2_exploitability": x_opencti_cvss_v2_exploitability,
                        "x_opencti_cvss_v2_remediation_level": x_opencti_cvss_v2_remediation_level,
                        "x_opencti_cvss_v2_report_confidence": x_opencti_cvss_v2_report_confidence,
                        "x_opencti_cvss_v2_temporal_score": x_opencti_cvss_v2_temporal_score,
                        # CVSS 4
                        "x_opencti_cvss_v4_vector_string": x_opencti_cvss_v4_vector_string,
                        "x_opencti_cvss_v4_base_score": x_opencti_cvss_v4_base_score,
                        "x_opencti_cvss_v4_base_severity": x_opencti_cvss_v4_base_severity,
                        "x_opencti_cvss_v4_attack_vector": x_opencti_cvss_v4_attack_vector,
                        "x_opencti_cvss_v4_attack_complexity": x_opencti_cvss_v4_attack_complexity,
                        "x_opencti_cvss_v4_attack_requirements": x_opencti_cvss_v4_attack_requirements,
                        "x_opencti_cvss_v4_privileges_required": x_opencti_cvss_v4_privileges_required,
                        "x_opencti_cvss_v4_user_interaction": x_opencti_cvss_v4_user_interaction,
                        "x_opencti_cvss_v4_confidentiality_impact_v": x_opencti_cvss_v4_confidentiality_impact_v,
                        "x_opencti_cvss_v4_confidentiality_impact_s": x_opencti_cvss_v4_confidentiality_impact_s,
                        "x_opencti_cvss_v4_integrity_impact_v": x_opencti_cvss_v4_integrity_impact_v,
                        "x_opencti_cvss_v4_integrity_impact_s": x_opencti_cvss_v4_integrity_impact_s,
                        "x_opencti_cvss_v4_availability_impact_v": x_opencti_cvss_v4_availability_impact_v,
                        "x_opencti_cvss_v4_availability_impact_s": x_opencti_cvss_v4_availability_impact_s,
                        "x_opencti_cvss_v4_exploit_maturity": x_opencti_cvss_v4_exploit_maturity,
                        # Others
                        "x_opencti_cwe": x_opencti_cwe,
                        "x_opencti_cisa_kev": x_opencti_cisa_kev,
                        "x_opencti_epss_score": x_opencti_epss_score,
                        "x_opencti_epss_percentile": x_opencti_epss_percentile,
                        "x_opencti_score": x_opencti_score,
                        "x_opencti_first_seen_active": x_opencti_first_seen_active,
                        "x_opencti_stix_ids": x_opencti_stix_ids,
                        "x_opencti_workflow_id": x_opencti_workflow_id,
                        "x_opencti_modified_at": x_opencti_modified_at,
                        "update": update,
                        "files": files,
                        "filesMarkings": files_markings,
                        "noTriggerImport": no_trigger_import,
                        "embedded": embedded,
                        "upsertOperations": upsert_operations,
                    }
                },
            )
            return self.opencti.process_multiple_fields(
                result["data"]["vulnerabilityAdd"]
            )
        else:
            self.opencti.app_logger.error(
                "[opencti_vulnerability] Missing parameters: name"
            )
            return None

    def import_from_stix2(self, **kwargs):
        """Import a Vulnerability object from a STIX2 object.

        :param stixObject: the STIX2 Vulnerability object
        :type stixObject: dict
        :param extras: extra parameters including created_by_id, object_marking_ids, etc.
        :type extras: dict
        :param update: whether to update if the entity already exists
        :type update: bool
        :return: Vulnerability object
        :rtype: dict or None
        """
        stix_object = kwargs.get("stixObject", None)
        extras = kwargs.get("extras", {})
        update = kwargs.get("update", False)

        if stix_object is not None:
            # Search in extensions
            if "x_opencti_stix_ids" not in stix_object:
                stix_object["x_opencti_stix_ids"] = (
                    self.opencti.get_attribute_in_extension("stix_ids", stix_object)
                )
            if "x_opencti_granted_refs" not in stix_object:
                stix_object["x_opencti_granted_refs"] = (
                    self.opencti.get_attribute_in_extension("granted_refs", stix_object)
                )
            if "x_opencti_workflow_id" not in stix_object:
                stix_object["x_opencti_workflow_id"] = (
                    self.opencti.get_attribute_in_extension("workflow_id", stix_object)
                )
            # Backward compatibility
            if "x_opencti_base_score" in stix_object:
                stix_object["x_opencti_cvss_base_score"] = stix_object[
                    "x_opencti_base_score"
                ]
            if "x_opencti_base_severity" in stix_object:
                stix_object["x_opencti_cvss_base_severity"] = stix_object[
                    "x_opencti_base_severity"
                ]
            if "x_opencti_attack_vector" in stix_object:
                stix_object["x_opencti_cvss_attack_vector"] = stix_object[
                    "x_opencti_attack_vector"
                ]
            if "x_opencti_integrity_impact" in stix_object:
                stix_object["x_opencti_cvss_integrity_impact"] = stix_object[
                    "x_opencti_integrity_impact"
                ]
            if "x_opencti_availability_impact" in stix_object:
                stix_object["x_opencti_cvss_availability_impact"] = stix_object[
                    "x_opencti_availability_impact"
                ]
            if "x_opencti_confidentiality_impact" in stix_object:
                stix_object["x_opencti_cvss_confidentiality_impact"] = stix_object[
                    "x_opencti_confidentiality_impact"
                ]

            # Search in extensions
            if "x_opencti_aliases" not in stix_object:
                stix_object["x_opencti_aliases"] = (
                    self.opencti.get_attribute_in_extension("aliases", stix_object)
                )
            # CVSS3
            if "x_opencti_cvss_vector_string" not in stix_object:
                stix_object["x_opencti_cvss_vector_string"] = (
                    self.opencti.get_attribute_in_extension("cvss_vector", stix_object)
                )
            if "x_opencti_cvss_base_score" not in stix_object:
                stix_object["x_opencti_cvss_base_score"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_base_score", stix_object
                    )
                )
            if "x_opencti_cvss_base_severity" not in stix_object:
                stix_object["x_opencti_cvss_base_severity"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_base_severity", stix_object
                    )
                )
            if "x_opencti_cvss_attack_vector" not in stix_object:
                stix_object["x_opencti_cvss_attack_vector"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_attack_vector", stix_object
                    )
                )
            if "x_opencti_cvss_attack_complexity" not in stix_object:
                stix_object["x_opencti_cvss_attack_complexity"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_attack_complexity", stix_object
                    )
                )
            if "x_opencti_cvss_privileges_required" not in stix_object:
                stix_object["x_opencti_cvss_privileges_required"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_privileges_required", stix_object
                    )
                )
            if "x_opencti_cvss_user_interaction" not in stix_object:
                stix_object["x_opencti_cvss_user_interaction"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_user_interaction", stix_object
                    )
                )
            if "x_opencti_cvss_scope" not in stix_object:
                stix_object["x_opencti_cvss_scope"] = (
                    self.opencti.get_attribute_in_extension("cvss_scope", stix_object)
                )
            if "x_opencti_cvss_confidentiality_impact" not in stix_object:
                stix_object["x_opencti_cvss_confidentiality_impact"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_confidentiality_impact", stix_object
                    )
                )
            if "x_opencti_cvss_integrity_impact" not in stix_object:
                stix_object["x_opencti_cvss_integrity_impact"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_integrity_impact", stix_object
                    )
                )
            if "x_opencti_cvss_availability_impact" not in stix_object:
                stix_object["x_opencti_cvss_availability_impact"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_availability_impact", stix_object
                    )
                )
            if "x_opencti_cvss_exploit_code_maturity" not in stix_object:
                stix_object["x_opencti_cvss_exploit_code_maturity"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_exploit_code_maturity", stix_object
                    )
                )
            if "x_opencti_cvss_remediation_level" not in stix_object:
                stix_object["x_opencti_cvss_remediation_level"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_remediation_level", stix_object
                    )
                )
            if "x_opencti_cvss_report_confidence" not in stix_object:
                stix_object["x_opencti_cvss_report_confidence"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_report_confidence", stix_object
                    )
                )
            if "x_opencti_cvss_temporal_score" not in stix_object:
                stix_object["x_opencti_cvss_temporal_score"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_temporal_score", stix_object
                    )
                )

            # CVSS2
            if "x_opencti_cvss_v2_vector_string" not in stix_object:
                stix_object["x_opencti_cvss_v2_vector_string"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_vector", stix_object
                    )
                )
            if "x_opencti_cvss_v2_base_score" not in stix_object:
                stix_object["x_opencti_cvss_v2_base_score"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_base_score", stix_object
                    )
                )
            if "x_opencti_cvss_v2_access_vector" not in stix_object:
                stix_object["x_opencti_cvss_v2_access_vector"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_access_vector", stix_object
                    )
                )
            if "x_opencti_cvss_v2_access_complexity" not in stix_object:
                stix_object["x_opencti_cvss_v2_access_complexity"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_access_complexity", stix_object
                    )
                )
            if "x_opencti_cvss_v2_authentication" not in stix_object:
                stix_object["x_opencti_cvss_v2_authentication"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_authentication", stix_object
                    )
                )
            if "x_opencti_cvss_v2_confidentiality_impact" not in stix_object:
                stix_object["x_opencti_cvss_v2_confidentiality_impact"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_confidentiality_impact", stix_object
                    )
                )
            if "x_opencti_cvss_v2_integrity_impact" not in stix_object:
                stix_object["x_opencti_cvss_v2_integrity_impact"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_integrity_impact", stix_object
                    )
                )
            if "x_opencti_cvss_v2_availability_impact" not in stix_object:
                stix_object["x_opencti_cvss_v2_availability_impact"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_availability_impact", stix_object
                    )
                )
            if "x_opencti_cvss_v2_exploitability" not in stix_object:
                stix_object["x_opencti_cvss_v2_exploitability"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_exploitability", stix_object
                    )
                )
            if "x_opencti_cvss_v2_remediation_level" not in stix_object:
                stix_object["x_opencti_cvss_v2_remediation_level"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_remediation_level", stix_object
                    )
                )
            if "x_opencti_cvss_v2_report_confidence" not in stix_object:
                stix_object["x_opencti_cvss_v2_report_confidence"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_report_confidence", stix_object
                    )
                )
            if "x_opencti_cvss_v2_temporal_score" not in stix_object:
                stix_object["x_opencti_cvss_v2_temporal_score"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v2_temporal_score", stix_object
                    )
                )

            # CVSS4
            if "x_opencti_cvss_v4_vector_string" not in stix_object:
                stix_object["x_opencti_cvss_v4_vector_string"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_vector", stix_object
                    )
                )
            if "x_opencti_cvss_v4_base_score" not in stix_object:
                stix_object["x_opencti_cvss_v4_base_score"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_base_score", stix_object
                    )
                )
            if "x_opencti_cvss_v4_base_severity" not in stix_object:
                stix_object["x_opencti_cvss_v4_base_severity"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_base_severity", stix_object
                    )
                )
            if "x_opencti_cvss_v4_attack_vector" not in stix_object:
                stix_object["x_opencti_cvss_v4_attack_vector"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_attack_vector", stix_object
                    )
                )
            if "x_opencti_cvss_v4_attack_complexity" not in stix_object:
                stix_object["x_opencti_cvss_v4_attack_complexity"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_attack_complexity", stix_object
                    )
                )
            if "x_opencti_cvss_v4_attack_requirements" not in stix_object:
                stix_object["x_opencti_cvss_v4_attack_requirements"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_attack_requirements", stix_object
                    )
                )
            if "x_opencti_cvss_v4_privileges_required" not in stix_object:
                stix_object["x_opencti_cvss_v4_privileges_required"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_privileges_required", stix_object
                    )
                )
            if "x_opencti_cvss_v4_user_interaction" not in stix_object:
                stix_object["x_opencti_cvss_v4_user_interaction"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_user_interaction", stix_object
                    )
                )
            if "x_opencti_cvss_v4_confidentiality_impact_v" not in stix_object:
                stix_object["x_opencti_cvss_v4_confidentiality_impact_v"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_confidentiality_impact_v", stix_object
                    )
                )
            if "x_opencti_cvss_v4_confidentiality_impact_s" not in stix_object:
                stix_object["x_opencti_cvss_v4_confidentiality_impact_s"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_confidentiality_impact_s", stix_object
                    )
                )
            if "x_opencti_cvss_v4_integrity_impact_v" not in stix_object:
                stix_object["x_opencti_cvss_v4_integrity_impact_v"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_integrity_impact_v", stix_object
                    )
                )
            if "x_opencti_cvss_v4_integrity_impact_s" not in stix_object:
                stix_object["x_opencti_cvss_v4_integrity_impact_s"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_integrity_impact_s", stix_object
                    )
                )
            if "x_opencti_cvss_v4_availability_impact_v" not in stix_object:
                stix_object["x_opencti_cvss_v4_availability_impact_v"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_availability_impact_v", stix_object
                    )
                )
            if "x_opencti_cvss_v4_availability_impact_s" not in stix_object:
                stix_object["x_opencti_cvss_v4_availability_impact_s"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_availability_impact_s", stix_object
                    )
                )
            if "x_opencti_cvss_v4_exploit_maturity" not in stix_object:
                stix_object["x_opencti_cvss_v4_exploit_maturity"] = (
                    self.opencti.get_attribute_in_extension(
                        "cvss_v4_exploit_maturity", stix_object
                    )
                )

            # Others
            if "x_opencti_cwe" not in stix_object:
                stix_object["x_opencti_cwe"] = self.opencti.get_attribute_in_extension(
                    "cwe", stix_object
                )
            if "x_opencti_cisa_kev" not in stix_object:
                stix_object["x_opencti_cisa_kev"] = (
                    self.opencti.get_attribute_in_extension("cisa_kev", stix_object)
                )
            if "x_opencti_epss_score" not in stix_object:
                stix_object["x_opencti_epss_score"] = (
                    self.opencti.get_attribute_in_extension("epss_score", stix_object)
                )
            if "x_opencti_epss_percentile" not in stix_object:
                stix_object["x_opencti_epss_percentile"] = (
                    self.opencti.get_attribute_in_extension(
                        "epss_percentile", stix_object
                    )
                )
            if "x_opencti_score" not in stix_object:
                stix_object["x_opencti_score"] = (
                    self.opencti.get_attribute_in_extension("score", stix_object)
                )
            if "x_opencti_first_seen_active" not in stix_object:
                stix_object["x_opencti_first_seen_active"] = (
                    self.opencti.get_attribute_in_extension(
                        "first_seen_active", stix_object
                    )
                )
            if "x_opencti_modified_at" not in stix_object:
                stix_object["x_opencti_modified_at"] = (
                    self.opencti.get_attribute_in_extension("modified_at", stix_object)
                )
            if "opencti_upsert_operations" not in stix_object:
                stix_object["opencti_upsert_operations"] = (
                    self.opencti.get_attribute_in_extension(
                        "opencti_upsert_operations", stix_object
                    )
                )

            return self.create(
                stix_id=stix_object["id"],
                createdBy=(
                    extras["created_by_id"] if "created_by_id" in extras else None
                ),
                objectMarking=(
                    extras["object_marking_ids"]
                    if "object_marking_ids" in extras
                    else None
                ),
                objectLabel=(
                    extras["object_label_ids"] if "object_label_ids" in extras else None
                ),
                externalReferences=(
                    extras["external_references_ids"]
                    if "external_references_ids" in extras
                    else None
                ),
                revoked=stix_object["revoked"] if "revoked" in stix_object else None,
                confidence=(
                    stix_object["confidence"] if "confidence" in stix_object else None
                ),
                lang=stix_object["lang"] if "lang" in stix_object else None,
                created=stix_object["created"] if "created" in stix_object else None,
                modified=stix_object["modified"] if "modified" in stix_object else None,
                name=stix_object["name"],
                description=(
                    self.opencti.stix2.convert_markdown(stix_object["description"])
                    if "description" in stix_object
                    else None
                ),
                x_opencti_aliases=(
                    stix_object["x_opencti_aliases"]
                    if "x_opencti_aliases" in stix_object
                    else None
                ),
                x_opencti_stix_ids=(
                    stix_object["x_opencti_stix_ids"]
                    if "x_opencti_stix_ids" in stix_object
                    else None
                ),
                objectOrganization=(
                    stix_object["x_opencti_granted_refs"]
                    if "x_opencti_granted_refs" in stix_object
                    else None
                ),
                x_opencti_workflow_id=(
                    stix_object["x_opencti_workflow_id"]
                    if "x_opencti_workflow_id" in stix_object
                    else None
                ),
                x_opencti_modified_at=(
                    stix_object["x_opencti_modified_at"]
                    if "x_opencti_modified_at" in stix_object
                    else None
                ),
                # CVSS3
                x_opencti_cvss_vector_string=(
                    stix_object["x_opencti_cvss_vector_string"]
                    if "x_opencti_cvss_vector_string" in stix_object
                    else None
                ),
                x_opencti_cvss_base_score=(
                    stix_object["x_opencti_cvss_base_score"]
                    if "x_opencti_cvss_base_score" in stix_object
                    else None
                ),
                x_opencti_cvss_base_severity=(
                    stix_object["x_opencti_cvss_base_severity"]
                    if "x_opencti_cvss_base_severity" in stix_object
                    else None
                ),
                x_opencti_cvss_attack_vector=(
                    stix_object["x_opencti_cvss_attack_vector"]
                    if "x_opencti_cvss_attack_vector" in stix_object
                    else None
                ),
                x_opencti_cvss_attack_complexity=(
                    stix_object["x_opencti_cvss_attack_complexity"]
                    if "x_opencti_cvss_attack_complexity" in stix_object
                    else None
                ),
                x_opencti_cvss_privileges_required=(
                    stix_object["x_opencti_cvss_privileges_required"]
                    if "x_opencti_cvss_privileges_required" in stix_object
                    else None
                ),
                x_opencti_cvss_user_interaction=(
                    stix_object["x_opencti_cvss_user_interaction"]
                    if "x_opencti_cvss_user_interaction" in stix_object
                    else None
                ),
                x_opencti_cvss_scope=(
                    stix_object["x_opencti_cvss_scope"]
                    if "x_opencti_cvss_scope" in stix_object
                    else None
                ),
                x_opencti_cvss_confidentiality_impact=(
                    stix_object["x_opencti_cvss_confidentiality_impact"]
                    if "x_opencti_cvss_confidentiality_impact" in stix_object
                    else None
                ),
                x_opencti_cvss_integrity_impact=(
                    stix_object["x_opencti_cvss_integrity_impact"]
                    if "x_opencti_cvss_integrity_impact" in stix_object
                    else None
                ),
                x_opencti_cvss_availability_impact=(
                    stix_object["x_opencti_cvss_availability_impact"]
                    if "x_opencti_cvss_availability_impact" in stix_object
                    else None
                ),
                x_opencti_cvss_exploit_code_maturity=(
                    stix_object["x_opencti_cvss_exploit_code_maturity"]
                    if "x_opencti_cvss_exploit_code_maturity" in stix_object
                    else None
                ),
                x_opencti_cvss_remediation_level=(
                    stix_object["x_opencti_cvss_remediation_level"]
                    if "x_opencti_cvss_remediation_level" in stix_object
                    else None
                ),
                x_opencti_cvss_report_confidence=(
                    stix_object["x_opencti_cvss_report_confidence"]
                    if "x_opencti_cvss_report_confidence" in stix_object
                    else None
                ),
                x_opencti_cvss_temporal_score=(
                    stix_object["x_opencti_cvss_temporal_score"]
                    if "x_opencti_cvss_temporal_score" in stix_object
                    else None
                ),
                # CVSS2
                x_opencti_cvss_v2_vector_string=(
                    stix_object["x_opencti_cvss_v2_vector_string"]
                    if "x_opencti_cvss_v2_vector_string" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_base_score=(
                    stix_object["x_opencti_cvss_v2_base_score"]
                    if "x_opencti_cvss_v2_base_score" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_access_vector=(
                    stix_object["x_opencti_cvss_v2_access_vector"]
                    if "x_opencti_cvss_v2_access_vector" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_access_complexity=(
                    stix_object["x_opencti_cvss_v2_access_complexity"]
                    if "x_opencti_cvss_v2_access_complexity" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_authentication=(
                    stix_object["x_opencti_cvss_v2_authentication"]
                    if "x_opencti_cvss_v2_authentication" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_confidentiality_impact=(
                    stix_object["x_opencti_cvss_v2_confidentiality_impact"]
                    if "x_opencti_cvss_v2_confidentiality_impact" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_integrity_impact=(
                    stix_object["x_opencti_cvss_v2_integrity_impact"]
                    if "x_opencti_cvss_v2_integrity_impact" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_availability_impact=(
                    stix_object["x_opencti_cvss_v2_availability_impact"]
                    if "x_opencti_cvss_v2_availability_impact" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_exploitability=(
                    stix_object["x_opencti_cvss_v2_exploitability"]
                    if "x_opencti_cvss_v2_exploitability" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_remediation_level=(
                    stix_object["x_opencti_cvss_v2_remediation_level"]
                    if "x_opencti_cvss_v2_remediation_level" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_report_confidence=(
                    stix_object["x_opencti_cvss_v2_report_confidence"]
                    if "x_opencti_cvss_v2_report_confidence" in stix_object
                    else None
                ),
                x_opencti_cvss_v2_temporal_score=(
                    stix_object["x_opencti_cvss_v2_temporal_score"]
                    if "x_opencti_cvss_v2_temporal_score" in stix_object
                    else None
                ),
                # CVSS4
                x_opencti_cvss_v4_vector_string=(
                    stix_object["x_opencti_cvss_v4_vector_string"]
                    if "x_opencti_cvss_v4_vector_string" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_base_score=(
                    stix_object["x_opencti_cvss_v4_base_score"]
                    if "x_opencti_cvss_v4_base_score" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_base_severity=(
                    stix_object["x_opencti_cvss_v4_base_severity"]
                    if "x_opencti_cvss_v4_base_severity" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_attack_vector=(
                    stix_object["x_opencti_cvss_v4_attack_vector"]
                    if "x_opencti_cvss_v4_attack_vector" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_attack_complexity=(
                    stix_object["x_opencti_cvss_v4_attack_complexity"]
                    if "x_opencti_cvss_v4_attack_complexity" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_attack_requirements=(
                    stix_object["x_opencti_cvss_v4_attack_requirements"]
                    if "x_opencti_cvss_v4_attack_requirements" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_privileges_required=(
                    stix_object["x_opencti_cvss_v4_privileges_required"]
                    if "x_opencti_cvss_v4_privileges_required" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_user_interaction=(
                    stix_object["x_opencti_cvss_v4_user_interaction"]
                    if "x_opencti_cvss_v4_user_interaction" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_confidentiality_impact_v=(
                    stix_object["x_opencti_cvss_v4_confidentiality_impact_v"]
                    if "x_opencti_cvss_v4_confidentiality_impact_v" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_confidentiality_impact_s=(
                    stix_object["x_opencti_cvss_v4_confidentiality_impact_s"]
                    if "x_opencti_cvss_v4_confidentiality_impact_s" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_integrity_impact_v=(
                    stix_object["x_opencti_cvss_v4_integrity_impact_v"]
                    if "x_opencti_cvss_v4_integrity_impact_v" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_integrity_impact_s=(
                    stix_object["x_opencti_cvss_v4_integrity_impact_s"]
                    if "x_opencti_cvss_v4_integrity_impact_s" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_availability_impact_v=(
                    stix_object["x_opencti_cvss_v4_availability_impact_v"]
                    if "x_opencti_cvss_v4_availability_impact_v" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_availability_impact_s=(
                    stix_object["x_opencti_cvss_v4_availability_impact_s"]
                    if "x_opencti_cvss_v4_availability_impact_s" in stix_object
                    else None
                ),
                x_opencti_cvss_v4_exploit_maturity=(
                    stix_object["x_opencti_cvss_v4_exploit_maturity"]
                    if "x_opencti_cvss_v4_exploit_maturity" in stix_object
                    else None
                ),
                # Others
                x_opencti_cwe=(
                    stix_object["x_opencti_cwe"]
                    if "x_opencti_cwe" in stix_object
                    else None
                ),
                x_opencti_cisa_kev=(
                    stix_object["x_opencti_cisa_kev"]
                    if "x_opencti_cisa_kev" in stix_object
                    else None
                ),
                x_opencti_epss_score=(
                    stix_object["x_opencti_epss_score"]
                    if "x_opencti_epss_score" in stix_object
                    else None
                ),
                x_opencti_epss_percentile=(
                    stix_object["x_opencti_epss_percentile"]
                    if "x_opencti_epss_percentile" in stix_object
                    else None
                ),
                x_opencti_score=(
                    stix_object["x_opencti_score"]
                    if "x_opencti_score" in stix_object
                    else None
                ),
                x_opencti_first_seen_active=(
                    stix_object["x_opencti_first_seen_active"]
                    if "x_opencti_first_seen_active" in stix_object
                    else None
                ),
                update=update,
                files=extras.get("files"),
                filesMarkings=extras.get("filesMarkings"),
                noTriggerImport=extras.get("noTriggerImport", None),
                embedded=extras.get("embedded", None),
                upsert_operations=(
                    stix_object["opencti_upsert_operations"]
                    if "opencti_upsert_operations" in stix_object
                    else None
                ),
            )
        else:
            self.opencti.app_logger.error(
                "[opencti_vulnerability] Missing parameters: stixObject"
            )
            return None



class OpenCTIApiInternalFile:
    """OpenCTI Internal File API class.

    Manages internal file operations.

    :param api: instance of :py:class:`~pycti.api.opencti_api_client.OpenCTIApiClient`
    :type api: OpenCTIApiClient
    """

    def __init__(self, api):
        """Initialize the OpenCTIApiInternalFile instance.

        :param api: OpenCTI API client instance
        :type api: OpenCTIApiClient
        """
        self.api = api

    def delete(self, **kwargs):
        """Delete an internal file.

        :param item: the item containing file information with id in extensions
        :type item: dict
        :return: None
        :rtype: None
        """
        item = kwargs.get("item", None)
        file_name = self.api.get_attribute_in_extension("id", item)
        if file_name is not None:
            query = """
                mutation InternalFileDelete($fileName: String) {
                    deleteImport(fileName: $fileName)
                }
               """
            self.api.query(
                query,
                {
                    "fileName": file_name,
                },
            )
        else:
            self.api.app_logger.error(
                "[opencti_internal_file] Cannot delete internal file, missing parameters: fileName"
            )
            return None



SCO_PROPERTIES = """
    id
    standard_id
    entity_type
    parent_types
    spec_version
    created_at
    updated_at
    objectOrganization {
        id
        standard_id
        name
    }
    creators {
        id
        name
    }
    createdBy {
        ... on Identity {
            id
            standard_id
            entity_type
            parent_types
            spec_version
            identity_class
            name
            description
            roles
            contact_information
            x_opencti_aliases
            created
            modified
            objectLabel {
                id
                value
                color
            }
        }
        ... on Organization {
            x_opencti_organization_type
            x_opencti_reliability
        }
        ... on Individual {
            x_opencti_firstname
            x_opencti_lastname
        }
    }
    objectMarking {
        id
        standard_id
        entity_type
        definition_type
        definition
        created
        modified
        x_opencti_order
        x_opencti_color
    }
    objectLabel {
        id
        value
        color
    }
    externalReferences {
        edges {
            node {
                id
                standard_id
                entity_type
                source_name
                description
                url
                hash
                external_id
                created
                modified
            }
        }
    }
    observable_value
    x_opencti_description
    x_opencti_score
    indicators {
        edges {
            node {
                id
                pattern
                pattern_type
            }
        }
    }
    ... on AutonomousSystem {
        number
        name
        rir
    }
    ... on Directory {
        path
        path_enc
        ctime
        mtime
        atime
    }
    ... on DomainName {
        value
    }
    ... on EmailAddr {
        value
        display_name
    }
    ... on EmailMessage {
        is_multipart
        attribute_date
        content_type
        message_id
        subject
        received_lines
        body
    }
    ... on Artifact {
        mime_type
        payload_bin
        url
        encryption_algorithm
        decryption_key
        hashes {
            algorithm
            hash
        }
        importFiles {
            edges {
                node {
                    id
                    name
                    size
                    metaData {
                        mimetype
                        version
                    }
                }
            }
        }
    }
    ... on StixFile {
        extensions
        size
        name
        name_enc
        magic_number_hex
        mime_type
        ctime
        mtime
        atime
        x_opencti_additional_names
        hashes {
          algorithm
          hash
        }
    }
    ... on X509Certificate {
        is_self_signed
        version
        serial_number
        signature_algorithm
        issuer
        subject
        subject_public_key_algorithm
        subject_public_key_modulus
        subject_public_key_exponent
        validity_not_before
        validity_not_after
        hashes {
          algorithm
          hash
        }
        basic_constraints
        name_constraints
        policy_constraints
        key_usage
        extended_key_usage
        subject_key_identifier
        authority_key_identifier
        subject_alternative_name
        issuer_alternative_name
        subject_directory_attributes
        crl_distribution_points
        inhibit_any_policy
        private_key_usage_period_not_before
        private_key_usage_period_not_after
        certificate_policies
        policy_mappings
    }
     ... on SSHKey {
        key_type
        public_key
        fingerprint_sha256
        fingerprint_md5
        key_length
        expiration_date
        comment
        created
    }
    ... on AIPrompt {
        value
    }
    ... on IPv4Addr {
        value
    }
    ... on IPv6Addr {
        value
    }
    ... on MacAddr {
        value
    }
    ... on Mutex {
        name
    }
    ... on NetworkTraffic {
        extensions
        start
        end
        is_active
        src_port
        dst_port
        protocols
        src_byte_count
        dst_byte_count
        src_packets
        dst_packets
    }
    ... on Process {
        extensions
        is_hidden
        pid
        created_time
        cwd
        command_line
        environment_variables
    }
    ... on Software {
        name
        cpe
        swid
        languages
        vendor
        version
        x_opencti_product
    }
    ... on Url {
        value
    }
    ... on UserAccount {
        extensions
        user_id
        credential
        account_login
        account_type
        display_name
        is_service_account
        is_privileged
        can_escalate_privs
        is_disabled
        account_created
        account_expires
        credential_last_changed
        account_first_login
        account_last_login
    }
    ... on WindowsRegistryKey {
        attribute_key
        modified_time
        number_of_subkeys
    }
    ... on WindowsRegistryValueType {
        name
        data
        data_type
    }
    ... on CryptographicKey {
        value
    }
    ... on CryptocurrencyWallet {
        value
    }
    ... on Hostname {
        value
    }
    ... on Text {
        value
    }
    ... on UserAgent {
        value
    }
    ... on BankAccount {
        iban
        bic
        account_number
    }
    ... on PhoneNumber {
        value
    }
    ... on TrackingNumber {
        value
    }
    ... on Credential {
        value
    }
    ... on PaymentCard {
        card_number
        expiration_date
        cvv
        holder_name
    }
    ... on Persona {
        persona_name
        persona_type
    }
    ... on MediaContent {
        title
        content
        media_category
        url
        publication_date
    }
"""
SCO_PROPERTIES_WITH_FILES = """
    id
    standard_id
    entity_type
    parent_types
    spec_version
    created_at
    updated_at
    objectOrganization {
        id
        standard_id
        name
    }
    creators {
        id
        name
    }
    createdBy {
        ... on Identity {
            id
            standard_id
            entity_type
            parent_types
            spec_version
            identity_class
            name
            description
            roles
            contact_information
            x_opencti_aliases
            created
            modified
            objectLabel {
                id
                value
                color
            }
        }
        ... on Organization {
            x_opencti_organization_type
            x_opencti_reliability
        }
        ... on Individual {
            x_opencti_firstname
            x_opencti_lastname
        }
    }
    objectMarking {
        id
        standard_id
        entity_type
        definition_type
        definition
        created
        modified
        x_opencti_order
        x_opencti_color
    }
    objectLabel {
        id
        value
        color
    }
    externalReferences {
        edges {
            node {
                id
                standard_id
                entity_type
                source_name
                description
                url
                hash
                external_id
                created
                modified
                importFiles {
                    edges {
                        node {
                            id
                            name
                            size
                            metaData {
                                mimetype
                                version
                            }
                        }
                    }
                }
            }
        }
    }
    observable_value
    x_opencti_description
    x_opencti_score
    indicators {
        edges {
            node {
                id
                pattern
                pattern_type
            }
        }
    }
    ... on AutonomousSystem {
        number
        name
        rir
    }
    ... on Directory {
        path
        path_enc
        ctime
        mtime
        atime
    }
    ... on DomainName {
        value
    }
    ... on EmailAddr {
        value
        display_name
    }
    ... on EmailMessage {
        is_multipart
        attribute_date
        content_type
        message_id
        subject
        received_lines
        body
    }
    ... on Artifact {
        mime_type
        payload_bin
        url
        encryption_algorithm
        decryption_key
        hashes {
            algorithm
            hash
        }
        importFiles {
            edges {
                node {
                    id
                    name
                    size
                    metaData {
                        mimetype
                        version
                    }
                }
            }
        }
    }
    ... on StixFile {
        extensions
        size
        name
        name_enc
        magic_number_hex
        mime_type
        ctime
        mtime
        atime
        x_opencti_additional_names
        hashes {
          algorithm
          hash
        }
    }
    ... on X509Certificate {
        is_self_signed
        version
        serial_number
        signature_algorithm
        issuer
        subject
        subject_public_key_algorithm
        subject_public_key_modulus
        subject_public_key_exponent
        validity_not_before
        validity_not_after
        hashes {
          algorithm
          hash
        }
        basic_constraints
        name_constraints
        policy_constraints
        key_usage
        extended_key_usage
        subject_key_identifier
        authority_key_identifier
        subject_alternative_name
        issuer_alternative_name
        subject_directory_attributes
        crl_distribution_points
        inhibit_any_policy
        private_key_usage_period_not_before
        private_key_usage_period_not_after
        certificate_policies
        policy_mappings
    }
    ... on SSHKey {
        key_type
        public_key
        fingerprint_sha256
        fingerprint_md5
        key_length
        expiration_date
        comment
        created
    }
    ... on AIPrompt {
        value
    }
    ... on IPv4Addr {
        value
    }
    ... on IPv6Addr {
        value
    }
    ... on MacAddr {
        value
    }
    ... on Mutex {
        name
    }
    ... on NetworkTraffic {
        extensions
        start
        end
        is_active
        src_port
        dst_port
        protocols
        src_byte_count
        dst_byte_count
        src_packets
        dst_packets
    }
    ... on Process {
        extensions
        is_hidden
        pid
        created_time
        cwd
        command_line
        environment_variables
    }
    ... on Software {
        name
        cpe
        swid
        languages
        vendor
        version
        x_opencti_product
    }
    ... on Url {
        value
    }
    ... on UserAccount {
        extensions
        user_id
        credential
        account_login
        account_type
        display_name
        is_service_account
        is_privileged
        can_escalate_privs
        is_disabled
        account_created
        account_expires
        credential_last_changed
        account_first_login
        account_last_login
    }
    ... on WindowsRegistryKey {
        attribute_key
        modified_time
        number_of_subkeys
    }
    ... on WindowsRegistryValueType {
        name
        data
        data_type
    }
    ... on CryptographicKey {
        value
    }
    ... on CryptocurrencyWallet {
        value
    }
    ... on Hostname {
        value
    }
    ... on Text {
        value
    }
    ... on UserAgent {
        value
    }
    ... on BankAccount {
        iban
        bic
        account_number
    }
    ... on PhoneNumber {
        value
    }
    ... on TrackingNumber {
        value
    }
    ... on Credential {
        value
    }
    ... on PaymentCard {
        card_number
        expiration_date
        cvv
        holder_name
    }
    ... on Persona {
        persona_name
        persona_type
    }
    ... on MediaContent {
        title
        content
        media_category
        url
        publication_date
    }
    importFiles {
        edges {
            node {
                id
                name
                size
                metaData {
                    mimetype
                    version
                }
                objectMarking {
                    id
                    standard_id
                    entity_type
                    definition_type
                    definition
                    created
                    modified
                    x_opencti_order
                    x_opencti_color
                }
            }
        }
    }
"""



from typing import Dict, List, Optional


class Role:
    """Representation of a role in OpenCTI

    Roles can have capabilities. Groups have roles, and the combined
    capabilities of those roles determine what a group of users can do on the
    platform.

    Check the properties attribute of the class to understand what default
    properties are fetched.

    :param opencti: instance of :py:class:`~pycti.api.opencti_api_client.OpenCTIApiClient`
    :type opencti: OpenCTIApiClient
    """

    def __init__(self, opencti):
        """Initialize the Role instance.

        :param opencti: OpenCTI API client instance
        :type opencti: OpenCTIApiClient
        """
        self.opencti = opencti
        self.properties = """
            id
            standard_id
            entity_type
            parent_types
            name
            description
            created_at
            updated_at
            capabilities {
                id
                name
                description
            }
            capabilitiesInDraft {
                id
                name
                description
            }
            can_manage_sensitive_config
        """

    def list(self, **kwargs) -> List[Dict]:
        """Search or list the roles on the server.

        :param search:
            Defaults to None.
        :type search: str, optional
        :param first: Defaults to 500 Return the first x results from ID or
            beginning if $after is not specified.
        :type first: int, optional
        :param after: Return all results after the given ID, useful for
            pagination. Ignored if returning all results, defaults to None.
        :type after: str, optional
        :param orderBy: Field to order by. Must be one of "name",
            "created_at", "updated_at", or "_score". Defaults
            to "name", defaults to "name".
        :type orderBy: str, optional
        :param orderMode: Direction to order in, either "asc" or "desc",
            defaults to "asc".
        :type orderMode: str, optional
        :param customAttributes: Defaults to None. Custom attributes to return
            from query. If None, defaults are used.
        :type customAttributes: str, optional
        :param getAll: Defaults to False. Retrieve all results. If true then
            the "first" param is ignored.
        :type getAll: bool, optional
        :param withPagination: Defaults to False Whether to include pagination
            pageInfo properties in result.
        :type withPagination: bool, optional

        :return: List of Python dictionaries with the properties of the role.
        :rtype: List[Dict]
        """
        search = kwargs.get("search", None)
        first = kwargs.get("first", 500)
        after = kwargs.get("after", None)
        order_by = kwargs.get("orderBy", None)
        order_mode = kwargs.get("orderMode", None)
        custom_attributes = kwargs.get("customAttributes", None)
        get_all = kwargs.get("getAll", False)
        with_pagination = kwargs.get("withPagination", False)

        self.opencti.admin_logger.info(
            "Searching roles matching search term", {"search": search}
        )

        if get_all:
            first = 100

        query = (
            """
            query RoleList($first: Int, $after: ID, $orderBy: RolesOrdering, $orderMode: OrderingMode, $search: String) {
                roles(first: $first, after: $after, orderBy: $orderBy, orderMode: $orderMode, search: $search) {
                    edges {
                        node {
                            """
            + (self.properties if custom_attributes is None else custom_attributes)
            + """
                        }
                    }
                    pageInfo {
                        startCursor
                        endCursor
                        hasNextPage
                        hasPreviousPage
                        globalCount
                    }
                }
            }
            """
        )
        result = self.opencti.query(
            query,
            {
                "first": first,
                "after": after,
                "orderBy": order_by,
                "orderMode": order_mode,
                "search": search,
            },
        )
        if get_all:
            final_data = []
            data = self.opencti.process_multiple(result["data"]["roles"])
            final_data = final_data + data
            while result["data"]["roles"]["pageInfo"]["hasNextPage"]:
                after = result["data"]["roles"]["pageInfo"]["endCursor"]
                result = self.opencti.query(
                    query,
                    {
                        "first": first,
                        "after": after,
                        "orderBy": order_by,
                        "orderMode": order_mode,
                        "search": search,
                    },
                )
                data = self.opencti.process_multiple(result["data"]["roles"])
                final_data = final_data + data
            return final_data
        else:
            return self.opencti.process_multiple(
                result["data"]["roles"], with_pagination
            )

    def read(self, **kwargs) -> Optional[Dict]:
        """Get a role given its ID or a search term

        One of id or search must be provided.

        :param id: ID of the role on the platform
        :type id: str, optional
        :param search: Search term for a role, e.g. its name
        :type search: str, optional
        :param customAttributes: Custom attributes on the role to return
        :type customAttributes: str, optional

        :return: Representation of the role
        :rtype: Optional[Dict]
        """
        id = kwargs.get("id", None)
        search = kwargs.get("search", None)
        custom_attributes = kwargs.get("customAttributes", None)

        if id is not None:
            self.opencti.admin_logger.info("Reading role", {"id": id})
            query = (
                """
                query RoleRead($id: String!) {
                    role(id: $id) {
                        """
                + (self.properties if custom_attributes is None else custom_attributes)
                + """
                    }
                }
                """
            )
            result = self.opencti.query(query, {"id": id})
            return self.opencti.process_multiple_fields(result["data"]["role"])
        elif search is not None:
            result = self.list(search=search)
            return result[0] if len(result) > 0 else None
        else:
            self.opencti.admin_logger.error(
                "[opencti_role] Missing parameters: id or search"
            )
            return None

    def delete(self, **kwargs):
        """Delete a role given its ID

        :param id: ID for the role on the platform.
        :type id: str
        """
        id = kwargs.get("id", None)

        if id is None:
            self.opencti.admin_logger.error("[opencti_role] Missing parameter: id")
            return None

        self.opencti.admin_logger.info("Deleting role", {"id": id})
        query = """
            mutation RoleDelete($id: ID!) {
                roleEdit(id: $id) {
                    delete
                }
            }
        """
        self.opencti.query(query, {"id": id})

    def create(self, **kwargs) -> Optional[Dict]:
        """Add a new role to OpenCTI.

        :param name: Name to assign to the role.
        :type name: str
        :param description: Optional. Description of the role, defaults to
            None.
        :type description: str, optional
        :param customAttributes: Custom attributes to return on role
        :type customAttributes: str, optional
        :return: Representation of the role.
        :rtype: Optional[Dict]
        """
        name = kwargs.get("name", None)
        description = kwargs.get("description", None)
        custom_attributes = kwargs.get("customAttributes", None)

        if name is None:
            self.opencti.admin_logger.error("[opencti_role] Missing parameter: name")
            return None

        self.opencti.admin_logger.info(
            "Creating new role", {"name": name, "description": description}
        )
        query = (
            """
            mutation RoleCreate($input: RoleAddInput!) {
                roleAdd(input: $input) {
                    """
            + (self.properties if custom_attributes is None else custom_attributes)
            + """
                }
            }
            """
        )
        result = self.opencti.query(
            query, {"input": {"name": name, "description": description}}
        )
        return self.opencti.process_multiple_fields(result["data"]["roleAdd"])

    def update_field(self, **kwargs) -> Optional[Dict]:
        """Updates a given role with the given inputs

        Example of input::

            [
                {
                    "key": "name",
                    "value": "NewCustomRole"
                },
                {
                    "key": "can_manage_sensitive_config",
                    "value": False
                }
            ]

        :param id: ID for the role on the platform
        :type id: str
        :param input: List of EditInput objects
        :type input: List[Dict]
        :param customAttributes: Custom attributes to return on the role
        :type customAttributes: str, optional

        :return: Representation of the role
        :rtype: Optional[Dict]
        """
        id = kwargs.get("id", None)
        input = kwargs.get("input", None)
        custom_attributes = kwargs.get("customAttributes", None)

        if id is None or input is None:
            self.opencti.admin_logger.error(
                "[opencti_role] Missing parameters: id and input"
            )
            return None

        self.opencti.admin_logger.info(
            "Editing role with input", {"id": id, "input": input}
        )
        query = (
            """
            mutation RoleUpdate($id: ID!, $input: [EditInput]!) {
                roleEdit(id: $id) {
                    fieldPatch(input: $input) {
                        """
            + (self.properties if custom_attributes is None else custom_attributes)
            + """
                    }
                }
            }
            """
        )
        result = self.opencti.query(query, {"id": id, "input": input})
        return self.opencti.process_multiple_fields(
            result["data"]["roleEdit"]["fieldPatch"]
        )

    def add_capability(self, **kwargs) -> Optional[Dict]:
        """Adds a capability to a role

        :param id: ID of the role.
        :type id: str
        :param capability_id: ID of the capability to add.
        :type capability_id: str
        :return: Representation of the relationship, including the role and
            capability
        :rtype: Optional[Dict]
        """
        id = kwargs.get("id", None)
        capability_id = kwargs.get("capability_id", None)

        if id is None or capability_id is None:
            self.opencti.admin_logger.error(
                "[opencti_role] Missing parameters: id and capability_id"
            )
            return None

        self.opencti.admin_logger.info(
            "Adding capability to role", {"roleId": id, "capabilityId": capability_id}
        )
        query = (
            """
            mutation RoleEditAddCapability($id: ID!, $input: InternalRelationshipAddInput!) {
                roleEdit(id: $id) {
                    relationAdd(input: $input) {
                        id
                        entity_type
                        parent_types
                        created_at
                        updated_at
                        from {
                            ... on Role {
                                """
            + self.properties
            + """
                            }
                        }
                        to {
                            ... on Capability {
                                id, name, description
                            }
                        }
                    }
                }
            }
            """
        )
        result = self.opencti.query(
            query,
            {
                "id": id,
                "input": {"relationship_type": "has-capability", "toId": capability_id},
            },
        )
        return self.opencti.process_multiple_fields(
            result["data"]["roleEdit"]["relationAdd"]
        )

    def delete_capability(self, **kwargs) -> Optional[Dict]:
        """Removes a capability from a role

        :param id: ID of the role
        :type id: str
        :param capability_id: ID of the capability to remove
        :type capability_id: str
        :return: Representation of the role after removing the capability
        :rtype: Optional[Dict]
        """
        id = kwargs.get("id", None)
        capability_id = kwargs.get("capability_id", None)

        if id is None or capability_id is None:
            self.opencti.admin_logger.error(
                "[opencti_role] Missing parameters: id and capability_id"
            )
            return None

        self.opencti.admin_logger.info(
            "Removing capability from role",
            {"roleId": id, "capabilityId": capability_id},
        )
        query = (
            """
            mutation RoleEditDeleteCapability($id: ID!, $toId: StixRef!) {
                roleEdit(id: $id) {
                    relationDelete(toId: $toId, relationship_type: "has-capability") {
                        """
            + self.properties
            + """
                    }
                }
            }
            """
        )
        result = self.opencti.query(query, {"id": id, "toId": capability_id})
        return self.opencti.process_multiple_fields(
            result["data"]["roleEdit"]["relationDelete"]
        )

    def process_multiple_fields(self, data):
        """Process and normalize fields in role data.

        :param data: the role data dictionary to process
        :type data: dict
        :return: the processed role data with normalized fields
        :rtype: dict
        """
        if "capabilities" in data:
            data["capabilities"] = self.opencti.process_multiple(data["capabilities"])
            data["capabilitiesIds"] = self.opencti.process_multiple_ids(
                data["capabilities"]
            )
        return data



"""STIX2 utility functions and mappings for OpenCTI.

This module provides utility classes and constants for working with STIX2 objects
in OpenCTI, including type mappings, pattern generation, and object reference counting.
"""

from typing import Any, Dict

from stix2 import EqualityComparisonExpression, ObjectPath, ObservationExpression

# Aliases field constants
ALIASES_FIELD = "aliases"
X_OPENCTI_ALIASES_FIELD = "x_opencti_aliases"

SUPPORTED_INTERNAL_OBJECTS = [
    "user",
    "group",
    "capability",
    "role",
    "settings",
    "notification",
    "work",
    "trash",
    "draftworkspace",
    "playbook",
    "deleteoperation",
    "workspace",
    "publicdashboard",
]

STIX_META_OBJECTS = [
    "label",
    "vocabulary",
    "kill-chain-phase",
]

STIX_CORE_OBJECTS = [
    "attack-pattern",
    "campaign",
    "case-incident",
    "x-opencti-case-incident",
    "case-rfi",
    "x-opencti-case-rfi",
    "case-rft",
    "x-opencti-case-rft",
    "channel",
    "course-of-action",
    "data-component",
    "x-mitre-data-component",
    "data-source",
    "x-mitre-data-source",
    "event",
    "external-reference",
    "feedback",
    "x-opencti-feedback",
    "grouping",
    "identity",
    "incident",
    "indicator",
    "infrastructure",
    "intrusion-set",
    "language",
    "location",
    "malware",
    "malware-analysis",
    "marking-definition",
    "narrative",
    "note",
    "observed-data",
    "opinion",
    "report",
    "task",
    "x-opencti-task",
    "threat-actor",
    "tool",
    "vulnerability",
    "security-coverage",
]

SUPPORTED_STIX_ENTITY_OBJECTS = STIX_META_OBJECTS + STIX_CORE_OBJECTS

STIX_CYBER_OBSERVABLE_MAPPING = {
    "autonomous-system": "Autonomous-System",
    "directory": "Directory",
    "domain-name": "Domain-Name",
    "email-addr": "Email-Addr",
    "email-message": "Email-Message",
    "email-mime-part-type": "Email-Mime-Part-Type",
    "artifact": "Artifact",
    "file": "StixFile",
    "x509-certificate": "X509-Certificate",
    "ipv4-addr": "IPv4-Addr",
    "ipv6-addr": "IPv6-Addr",
    "mac-addr": "Mac-Addr",
    "mutex": "Mutex",
    "network-traffic": "Network-Traffic",
    "process": "Process",
    "software": "Software",
    "url": "Url",
    "user-account": "User-Account",
    "windows-registry-key": "Windows-Registry-Key",
    "windows-registry-value-type": "Windows-Registry-Value-Type",
    "hostname": "Hostname",
    "cryptographic-key": "Cryptographic-Key",
    "cryptocurrency-wallet": "Cryptocurrency-Wallet",
    "text": "Text",
    "user-agent": "User-Agent",
    "bank-account": "Bank-Account",
    "phone-number": "Phone-Number",
    "credential": "Credential",
    "tracking-number": "Tracking-Number",
    "payment-card": "Payment-Card",
    "media-content": "Media-Content",
    "simple-observable": "Simple-Observable",
    "persona": "Persona",
    "ssh-key": "SSH-Key",
    "ai-prompt": "AI-Prompt",
}

STIX_OBJECTS = (
    SUPPORTED_STIX_ENTITY_OBJECTS  # entities
    + list(STIX_CYBER_OBSERVABLE_MAPPING.keys())  # observables
    + ["relationship", "sighting"]  # relationships
)

PATTERN_MAPPING = {
    "Autonomous-System": ["number"],
    "Directory": ["path"],
    "Domain-Name": ["value"],
    "Email-Addr": ["value"],
    "File_md5": ["hashes", "MD5"],
    "File_sha1": ["hashes", "SHA-1"],
    "File_sha256": ["hashes", "SHA-256"],
    "File_sha512": ["hashes", "SHA-512"],
    "Email-Message_Body": ["body"],
    "Email-Message_Subject": ["subject"],
    "Email-Mime-Part-Type": ["body"],
    "IPv4-Addr": ["value"],
    "IPv6-Addr": ["value"],
    "Mac-Addr": ["value"],
    "Mutex": ["name"],
    "Network-Traffic": ["dst_port"],
    "Process": ["pid"],
    "Software": ["name"],
    "Url": ["value"],
    "User-Account": ["account_login"],
    "Windows-Registry-Key": ["key"],
    "Windows-Registry-Value-Type": ["name"],
    "Hostname": ["value"],
    "Bank-Account": ["iban"],
    "Phone-Number": ["value"],
    "Payment-Card": ["card_number"],
    "Tracking-Number": ["value"],
    "Credential": ["value"],
    "Media-Content": ["url"],
    "AI-Prompt": ["value"],
}

OBSERVABLES_VALUE_INT = [
    "Autonomous-System.number",
    "Network-Traffic.dst_port",
    "Process.pid",
]


class OpenCTIStix2Utils:
    """Utility class for STIX2 operations in OpenCTI.

    Provides helper methods for STIX2 conversions and pattern generation,
    including type mappings, observable pattern creation, and reference counting.
    """

    @staticmethod
    def stix_observable_opencti_type(observable_type):
        """Convert STIX observable type to OpenCTI type.

        :param observable_type: STIX observable type
        :type observable_type: str
        :return: Corresponding OpenCTI type or "Unknown"
        :rtype: str
        """
        if observable_type in STIX_CYBER_OBSERVABLE_MAPPING:
            return STIX_CYBER_OBSERVABLE_MAPPING[observable_type]
        else:
            return "Unknown"

    @staticmethod
    def create_stix_pattern(observable_type, observable_value):
        """Create a STIX pattern from an observable type and value.

        :param observable_type: Type of the observable
        :type observable_type: str
        :param observable_value: Value of the observable
        :type observable_value: str
        :return: STIX pattern string or None if type not supported
        :rtype: str or None
        """
        if observable_type in PATTERN_MAPPING:
            lhs = ObjectPath(
                (
                    observable_type.lower()
                    if "_" not in observable_type
                    else observable_type.split("_")[0].lower()
                ),
                PATTERN_MAPPING[observable_type],
            )
            ece = ObservationExpression(
                EqualityComparisonExpression(lhs, observable_value)
            )
            return str(ece)
        else:
            return None

    @staticmethod
    def generate_random_stix_id(stix_type):
        """Generate random stix id (uuid v1) - DEPRECATED.

        This function is deprecated and should not be used anymore.
        Please use the generate_id function for SDO or proper SCO constructor.

        :param stix_type: the stix type
        :raises ValueError: Always raises an error as this function is deprecated
        """
        raise ValueError(
            "This function should not be used anymore, please use the generate_id function for SDO or proper SCO constructor"
        )

    @staticmethod
    def retrieve_class_for_method(
        opencti_api_client, entity: Dict, type_path: str, method: str
    ) -> Any:
        """Retrieve the appropriate API class for a given entity type and method.

        :param opencti_api_client: OpenCTI API client instance
        :type opencti_api_client: OpenCTIApiClient
        :param entity: Entity dictionary containing the type
        :type entity: Dict
        :param type_path: Path to the type field in the entity
        :type type_path: str
        :param method: Name of the method to check for
        :type method: str
        :return: The API class that has the specified method, or None
        :rtype: Any
        """
        if entity is not None and type_path in entity:
            attribute_name = entity[type_path].lower().replace("-", "_")
            if hasattr(opencti_api_client, attribute_name):
                attribute = getattr(opencti_api_client, attribute_name)
                if hasattr(attribute, method):
                    return attribute
        return None

    @staticmethod
    def retrieveClassForMethod(
        openCTIApiClient, entity: Dict, type_path: str, method: str
    ) -> Any:
        """Retrieve the appropriate API class for a given entity type and method.

        .. deprecated::
            Use :meth:`retrieve_class_for_method` instead.

        :param openCTIApiClient: OpenCTI API client instance
        :type openCTIApiClient: OpenCTIApiClient
        :param entity: Entity dictionary containing the type
        :type entity: Dict
        :param type_path: Path to the type field in the entity
        :type type_path: str
        :param method: Name of the method to check for
        :type method: str
        :return: The API class that has the specified method, or None
        :rtype: Any
        """
        return OpenCTIStix2Utils.retrieve_class_for_method(
            openCTIApiClient, entity, type_path, method
        )

    @staticmethod
    def compute_object_refs_number(entity: Dict):
        """Compute the number of object references in an entity.

        :param entity: Entity dictionary to analyze
        :type entity: Dict
        :return: Total number of references
        :rtype: int
        """
        refs_number = 0
        for key in list(entity.keys()):
            if key.endswith("_refs") and entity[key] is not None:
                refs_number += len(entity[key])
            elif key.endswith("_ref"):
                refs_number += 1
            elif key == "external_references" and entity[key] is not None:
                refs_number += len(entity[key])
            elif key == "kill_chain_phases" and entity[key] is not None:
                refs_number += len(entity[key])
        return refs_number


# Types that use x_opencti_aliases instead of aliases
# Based on opencti-graphql/src/schema/stixDomainObject.ts resolveAliasesField()
_X_OPENCTI_ALIASES_TYPES = frozenset(
    ["course-of-action", "vulnerability", "grouping", "identity", "location"]
)

# Types that support aliases (from STIX_DOMAIN_OBJECT_ALIASED in stixDomainObject.ts)
_STIX_ALIASED_TYPES = frozenset(
    [
        "attack-pattern",
        "campaign",
        "channel",
        "x-opencti-channel",
        "course-of-action",
        "event",
        "x-opencti-event",
        "grouping",
        "identity",
        "incident",
        "infrastructure",
        "intrusion-set",
        "location",
        "malware",
        "narrative",
        "x-opencti-narrative",
        "threat-actor",
        "tool",
        "vulnerability",
    ]
)


def resolve_aliases_field(stix_type: str) -> str:
    """Resolve the correct aliases field name for a given STIX type.

    OpenCTI uses two different field names for aliases depending on the entity type:
    - `aliases`: Standard STIX field used by most SDO types (Attack-Pattern, Campaign,
      Infrastructure, Intrusion-Set, Malware, Threat-Actor-Group, Tool, Incident, etc.)
    - `x_opencti_aliases`: OpenCTI extension field used by Course-Of-Action, Vulnerability,
      Grouping, Identity types (Individual, Sector, System, Organization), and Location types
      (Region, Country, Administrative-Area, City, Position)

    This mirrors the logic in opencti-graphql/src/schema/stixDomainObject.ts resolveAliasesField()

    Note: This function is case-insensitive.

    :param stix_type: The STIX object type (e.g., "malware", "vulnerability", "identity")
    :type stix_type: str
    :return: The aliases field name to use ("aliases" or "x_opencti_aliases")
    :rtype: str

    Example:
        >>> resolve_aliases_field("malware")
        'aliases'
        >>> resolve_aliases_field("Vulnerability")
        'x_opencti_aliases'
        >>> resolve_aliases_field("IDENTITY")
        'x_opencti_aliases'
    """
    if stix_type.lower() in _X_OPENCTI_ALIASES_TYPES:
        return X_OPENCTI_ALIASES_FIELD
    return ALIASES_FIELD


def is_stix_object_aliased(stix_type: str) -> bool:
    """Check if a STIX object type supports aliases.

    Returns True for entity types that have an aliases field in OpenCTI.

    Note: This function is case-insensitive.

    :param stix_type: The STIX object type (e.g., "malware", "indicator", "identity")
    :type stix_type: str
    :return: True if the type supports aliases, False otherwise
    :rtype: bool

    Example:
        >>> is_stix_object_aliased("malware")
        True
        >>> is_stix_object_aliased("Malware")
        True
        >>> is_stix_object_aliased("indicator")
        False
    """
    return stix_type.lower() in _STIX_ALIASED_TYPES
